#!/usr/bin/env python2
from __future__ import print_function

"""filedrop.py: http upload target"""
__version__   = "1.2"
__author__    = "ed <filedrop@ocv.me>"
__url__       = "https://ocv.me/dev/?filedrop.py"
__credits__   = ["stackoverflow.com"]
__license__   = "MIT"
__copyright__ = 2019


# expect about 400 MiB/s, with py2 being slightly faster than py3
#
# navigate to your ip:port combination with a browser to see a
# basic upload form, or just use curl to upload windows7.iso:
#   curl -F "f=@windows7.iso" http://127.0.0.1:1234/
#
# the http response is a minimal html which contains the
# filesize and sha512 checksum of the uploaded file, as seen by us
#
# for a smaller alternative, see https://ocv.me/dev/?filesink.py
# which uses the built-in http parser (cgi.FieldStorage) and thus
# generates tempfiles and maxes out at 32 MiB/s (this one does 2300)
#
#-----------------------------------------------------------------------
#
# 2019-05-20, 00:09, v1.2
#  + allow multiple parallel transfers, see max_threads
#  + change from sha1 to truncated sha512 (slowdown to 76%)
#  + display primary local ip (default route) at startup
#
# 2019-05-13, 00:27, v1.1
#  + allow multiple files in one post (slowdown to 94%)
#
# 2019-05-01, 17:29, v1.0
#  + initial working release


import os
import sys
import time
import hashlib
import select
import socket
import logging
import threading


##
## CONFIG  (last definition wins)

bind_port = 1234 # argv[1] overrides this

bind_ip = '127.0.0.1' # only allow local connections
bind_ip = '0.0.0.0'   # allow connections from all interfaces

nullwrite = True  # disable all file output, for benchmark purposes
nullwrite = False # regular operation

max_threads = 4   # max parallel transfers
max_threads = 128 # max parallel transfers

## END CONFIG
##


logging.basicConfig(
	level = logging.DEBUG,
	format = '\033[36m%(asctime)s \033[33m%(levelname)5s \033[0m%(message)s'
)

PY2 = (sys.version_info[0] == 2)
if PY2:
	from Queue import Queue
else:
	from queue import Queue
	unicode = str


class Unrecv(object):
	def __init__(self, s):
		self.s = s
		self.buf = b''
	
	def recv(self, nbytes):
		if self.buf:
			ret = self.buf[:nbytes]
			self.buf = self.buf[nbytes:]
			return ret
		
		try:
			return self.s.recv(nbytes)
		except:
			return b''
	
	def unrecv(self, buf):
		self.buf = buf + self.buf


class UnbufferedHttpd(object):
	def __init__(self, addr, port):
		srv = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		srv.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		srv.bind((addr, port))
		srv.listen(max_threads)
		self.threads = Queue(max_threads)
		
		while True:
			self.threads.put(1)
			try:
				cli = srv.accept()
			except KeyboardInterrupt:
				print()
				logging.info('bye')
				sys.exit(0)
			
			thr = threading.Thread(target=self.client_thr, args=(cli,))
			thr.daemon = True
			thr.start()
	
	def client_thr(self, cli):
		try:
			UnbufferedClient(*cli).run()
		finally:
			self.threads.get()


class UnbufferedClient(object):
	def __init__(self, sck, addr):
		self.s = sck
		self.sr = Unrecv(sck)
		self.bufsz = 1024 * 32
		self.addr = addr
		self.ok = True
	
	def run(self):
		headerlines = self.read_header()
		if not self.ok:
			return
		
		self.headers = {}
		mode, self.req, _ = headerlines[0].split(' ')
		
		for header_line in headerlines[1:]:
			k, v = header_line.split(':', 1)
			self.headers[k.lower()] = v.strip()
		
		#self.bufsz = int(self.req.split('/')[-1]) * 1024
		
		if mode == 'GET':
			self.handle_get()
		elif mode == 'POST':
			self.handle_post()
		else:
			self.loud_reply(u'invalid HTTP mode "{0}"'.format(mode))
	
	def panic(self, msg):
		logging.error('client disconnected ({0})'.format(msg).upper())
		self.ok = False
		self.s.close()
	
	def read_header(self):
		ret = b''
		while True:
			if ret.endswith(b'\r\n\r\n'):
				break
			elif ret.endswith(b'\r\n\r'):
				n = 1
			elif ret.endswith(b'\r\n'):
				n = 2
			elif ret.endswith(b'\r'):
				n = 3
			else:
				n = 4
			
			buf = self.sr.recv(n)
			if not buf:
				self.panic('headers')
				break
			
			ret += buf
		
		return ret[:-4].decode('utf-8', 'replace').split('\r\n')
	
	def reply(self, body):
		header = 'HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nContent-Length: {0}\r\n\r\n'.format(len(body)).encode('utf-8')
		if self.ok:
			self.s.send(header + body)
		
		self.s.close()
		return body
	
	def loud_reply(self, body):
		logging.info(body.rstrip())
		self.reply(b'<pre>' + body.encode('utf-8'))
	
	def handle_get(self):
		logging.info('')
		logging.info('GET  {0} {1}'.format(self.addr[0], self.req))
		self.reply(b'<form method="post" enctype="multipart/form-data"><h5><input type="file" name="f" multiple><h5><input type="submit" value="start upload">')
	
	def handle_post(self):
		logging.info('')
		logging.info('POST {0} {1}'.format(self.addr[0], self.req))
		
		try:
			if self.headers['expect'].lower() == '100-continue':
				self.s.send(b'HTTP/1.1 100 Continue\r\n\r\n')
		except:
			pass
		
		form_segm = self.read_header()
		if not self.ok:
			return
		
		self.boundary = b'\r\n' + form_segm[0].encode('utf-8')
		for ln in form_segm[1:]:
			logging.debug(ln)
		
		fn = '/dev/null'
		fn0 = 'inc.{0:.6f}'.format(time.time())

		files = []
		t0 = time.time()
		for nfile in range(99):
			if not nullwrite:
				fn = '{0}.{1}'.format(fn0, nfile)
			
			with open(fn, 'wb') as f:
				logging.info('writing to {0}'.format(fn))
				sz, sha512 = self.handle_multipart(f)
				if sz == 0:
					break
				
				files.append([sz, sha512])
			
			buf = self.sr.recv(2)
			
			if buf == b'--':
				# end of multipart
				break
			
			if buf != b'\r\n':
				return self.loud_reply(u'protocol error')
			
			header = self.read_header()
			if not self.ok:
				break
			
			form_segm += header
			for ln in header:
				logging.debug(ln)
		
		td = time.time() - t0
		sz_total = sum(x[0] for x in files)
		spd = (sz_total / td) / (1024*1024)
		
		status = 'OK'
		if not self.ok:
			status = 'ERROR'
		
		msg = u'{0} // {1} bytes // {2:.3f} MiB/s\n'.format(status, sz_total, spd)
		
		for sz, sha512 in files:
			msg += u'sha512: {0} // {1} bytes\n'.format(sha512[:56], sz)
			# truncated SHA-512 prevents length extension attacks;
			# using SHA-512/224, optionally SHA-512/256 = :64
		
		self.loud_reply(msg)

		if not nullwrite:
			with open(fn0 + '.txt', 'wb') as f:
				f.write((u'\n'.join(unicode(x) for x in [
					u':'.join(unicode(x) for x in self.addr),
					u'\n'.join(form_segm),
					msg.rstrip()
					]) + '\n').encode('utf-8'))
	
	def handle_multipart(self, ofd):
		tlen = 0
		hashobj = hashlib.sha512()
		for buf in self.extract_filedata():
			tlen += len(buf)
			hashobj.update(buf)
			ofd.write(buf)
		
		return tlen, hashobj.hexdigest()
	
	def extract_filedata(self):
		blen = len(self.boundary)
		bufsz = self.bufsz
		while True:
			buf = self.sr.recv(bufsz)
			if not buf:
				# abort: client disconnected
				self.panic('outer')
				return
			
			while True:
				ofs = buf.find(self.boundary)
				if ofs != -1:
					self.sr.unrecv(buf[ofs + blen:])
					yield buf[:ofs]
					return
				
				d = len(buf) - blen
				if d > 0:
					# buffer growing large; yield everything except
					# the part at the end (maybe start of boundary)
					yield buf[:d]
					buf = buf[d:]
				
				# look for boundary near the end of the buffer
				for n in range(1, len(buf)+1):
					if not buf[-n:] in self.boundary:
						n -= 1
						break
				
				if n == 0 \
				or not self.boundary.startswith(buf[-n:]):
					# no boundary contents near the buffer edge
					break
				
				if blen == n:
					# EOF: found boundary
					yield buf[:-n]
					return
				
				buf2 = self.sr.recv(bufsz)
				if not buf2:
					# abort: client disconnected
					self.panic('inner')
					return
				
				buf += buf2
			
			yield buf


if __name__ == '__main__':
	if len(sys.argv) > 1:
		bind_port = int(sys.argv[1])

	ip = '127.0.0.1'
	if bind_ip != ip:
		s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
		try:
			s.connect(('10.255.255.255', 1))
			ip = s.getsockname()[0]
		except:
			pass
		s.close()

	logging.info('listening @ {0}:{1}'.format(bind_ip, bind_port))
	logging.info('available @ http://{0}:{1}/'.format(ip, bind_port))
	
	# this may look silly but...
	thr = threading.Thread(target=UnbufferedHttpd, args=(bind_ip, bind_port))
	thr.daemon = True
	thr.start()

	while True:
		time.sleep(60)
		# ...now it works on winxp :^)
